""" Spike detection using the multi-resolution 
Teager Energy Operator (MTEO).

Spike energy is calculated using the MTEO (with user supplied 
"k Values"). When the energy is above "Threshold quantile" times 
"Threshold factor", a period is considered a spike. Only the maximum 
of a period of length "Spike samples" gets a spike train entry.

If exactly one recording channel group is selected, a new unit named
"Detection result unit name" is created. If no recording channel group
is selected, the resulting spike trains are only attached to the
respective segment and no unit is created.

If there are artifact epochs that should not be considered for spike
detection, check "Ignore artifact epochs" and set "Artifact Tag" to 
the value of their 'tag' annotation.
"""
from spykeutils.plugin import analysis_plugin, gui_data
import scipy as sp
import neo
from PyQt4.QtGui import QMessageBox
from botmpy.nodes import SDMteoNode
from botmpy.common import merge_epochs

tag_prop = gui_data.ValueProp(False)

class SpikeDetectionPlugin(analysis_plugin.AnalysisPlugin):
    """ Spike Detection MTEO """
    quantile = gui_data.FloatItem('Threshold quantile', default=0.98,
                                  min=0.1, max=0.9999999)
    factor = gui_data.FloatItem('Threshold factor', default=1.0,
                                min=0.0)
    samples = gui_data.IntItem('Spike samples', default=49)
    k_values = gui_data.StringItem('k Values', default='3,9,15,21')
    unit_name = gui_data.StringItem('Detection result unit name', 
                                    default='MUA')
    ignore_artifacts = gui_data.BoolItem(
        'Remove artifact epochs').set_prop('display', store=tag_prop)
    artifact_tag = gui_data.StringItem(
        'Artifact tag').set_prop('display', active=tag_prop)
                                

    def get_name(self):
        return 'Spike Detection'

    def start(self, current, selections):
        if len(current.recording_channel_groups()) > 1:
            QMessageBox.critical(
                None, 
                'Too many recording channel groups selected',
                'Please select one recording channel group '
                '(or none if you want to detect spikes for '
                'signals only referenced in segments)')
            return
        
        self.detector = SDMteoNode(kvalues=self.parse_kvals(),
                                   quantile=self.quantile,
                                   threshold_factor=self.factor,
                                   tf=self.samples, 
                                   min_dist=self.samples / 2, 
                                   ch_separate=True)
            
        unit = neo.Unit(self.unit_name)
        rcg = None
        if len(current.recording_channel_groups()) == 1:
            unit.annotate(kvalues=str(self.parse_kvals()))
            unit.annotate(quantile=self.quantile)
            unit.annotate(threshold_factor=self.factor)
            unit.annotate(spike_samples=self.samples)
            rcg = current.recording_channel_groups()[0]
            
        current.progress.begin('Detecting spikes')
        data = current.analog_signals_by_segment(3)
        epochs = current.epochs()
        current.progress.set_ticks(len(data))
        
        for seg, sigs in data.iteritems():
            # Create data for spike detector
            sig = self.prepare_data(sigs)
                
            # Cut artifacts
            eps = []  # List of index tuples for non-artifact epochs
            arts = []  # List of artifact epochs
            if self.ignore_artifacts and seg in epochs:
                arts = [e for e in epochs[seg] 
                        if e.annotations.get('tag', None) == self.artifact_tag]
            
            sampling_rate = sigs[0].sampling_rate
            last_idx = 0
            for a in arts:
                eps.append((last_idx,
                            int((a.time * sampling_rate).simplified)))
                last_idx = int(
                    ((a.time + a.duration) * sampling_rate).simplified)
            eps.append((last_idx, sig.shape[0]))
            eps = merge_epochs(eps, min_dist=self.samples)
            
            # Use SDMteoNode to find spikes
            events = []
            for ep in eps:
                self.detector.reset()
                self.detector(sig[ep[0]:ep[1], :])
                events.extend(self.detector.events + ep[0])
            
            st = neo.SpikeTrain(events * sigs[0].sampling_period.simplified + 
                                sigs[0].t_start,
                                t_start=sigs[0].t_start, t_stop=sigs[0].t_stop)
            st.segment = seg
            if rcg:
                st.unit = unit
            unit.spiketrains.append(st)
           
            current.progress.step()

        if not unit.spiketrains:
            QMessageBox.warning(None, 'No spikes found',
                                'No spiketrains have been created. Maybe '
                                'there are no analog signals selected?')
        # Attach to object hierarchy
        for st in unit.spiketrains:
            st.segment.spiketrains.append(st)   
        if rcg:
            rcg.units.append(unit)
            unit.recordingchannelgroup = rcg
        
        current.progress.done()
        current.refresh_view()
        
    def prepare_data(self, sigs):
        try:
            return sp.vstack([sp.asarray(s) for s in sigs]).T
        except ValueError:
            segment_desc = ''
            if seg.name:
                segment_desc = 'in segment %s ' % seg.name
            raise SpykeException(
                'Error: Analog signals ' + segment_desc + 
                'do not have the same length!')
        
    def parse_kvals(self):
        val_strings = self.k_values.replace(' ', '').split(',')
        values = set()
        for sv in val_strings:
            v = int(sv)
            if not v % 2:
                raise ValueError('%d is not an odd value!' % v)
            values.add(v)
        return sorted(values)
            
    def configure(self):
        super(SpikeDetectionPlugin, self).configure()
        input_ok = False
        
        while not input_ok:
            try:
                vals = self.parse_kvals()
                input_ok = True
            except Exception, e:
                QMessageBox.warning(None, 'Unable to set parameters',
                    'k Values needs to be a comma separated string of '
                    'odd integers:\n%s' % str(e))
                super(SpikeDetectionPlugin, self).configure()